@@ -30,10 +30,9 @@ import time
 import warnings
 
 from libarchive import _libarchive
-try:
-    from cStringIO import StringIO
-except ImportError:
-    from StringIO import StringIO
+from io import StringIO
+
+PY3 = sys.version_info[0] == 3
 
 # Suggested block size for libarchive. Libarchive may adjust it.
 BLOCK_SIZE = 10240
@@ -134,7 +133,7 @@ def is_archive_name(filename, formats=None):
     This function will return the name of the most likely archive format, None if the file is
     unlikely to be an archive.'''
     if formats is None:
-        formats = FORMAT_EXTENSIONS.values()
+        formats = list(FORMAT_EXTENSIONS.values())
     format, filter = guess_format(filename)
     if format in formats:
         return format
@@ -153,8 +152,8 @@ def is_archive(f, formats=(None, ), filters=(None, )):
 
     This function will return True if the file can be opened as an archive using the given
     format(s)/filter(s).'''
-    if isinstance(f, basestring):
-        f = file(f, 'r')
+    if isinstance(f, str):
+        f = open(f, 'r')
     a = _libarchive.archive_read_new()
     for format in formats:
         format = get_func(format, FORMATS, 0)
@@ -175,6 +174,7 @@ def is_archive(f, formats=(None, ), filters=(None, )):
     finally:
         _libarchive.archive_read_close(a)
         _libarchive.archive_read_free(a)
+        f.close()
 
 
 class EntryReadStream(object):
@@ -271,7 +271,7 @@ class EntryWriteStream(object):
         if self.buffer:
             self.buffer.write(data)
         else:
-            _libarchive.archive_write_data_from_str(self.archive._a, data)
+            _libarchive.archive_write_data_from_str(self.archive._a, data.encode('utf-8'))
         self.bytes += len(data)
 
     def close(self):
@@ -280,7 +280,7 @@ class EntryWriteStream(object):
         if self.buffer:
             self.entry.size = self.buffer.tell()
             self.entry.to_archive(self.archive)
-            _libarchive.archive_write_data_from_str(self.archive._a, self.buffer.getvalue())
+            _libarchive.archive_write_data_from_str(self.archive._a, self.buffer.getvalue().encode('utf-8'))
         _libarchive.archive_write_finish_entry(self.archive._a)
 
         # Call archive.close() with _defer True to let it know we have been
@@ -312,8 +312,13 @@ class Entry(object):
             call_and_check(_libarchive.archive_read_next_header2, archive._a, archive._a, e)
             mode = _libarchive.archive_entry_filetype(e)
             mode |= _libarchive.archive_entry_perm(e)
-            entry = cls(
+            if PY3:
+                pathname=_libarchive.archive_entry_pathname(e)
+            else:
                 pathname=_libarchive.archive_entry_pathname(e).decode(encoding),
+
+            entry = cls(
+                pathname=pathname,
                 size=_libarchive.archive_entry_size(e),
                 mtime=_libarchive.archive_entry_mtime(e),
                 mode=mode,
@@ -330,7 +335,7 @@ class Entry(object):
         if entry is None:
             entry = cls(encoding=encoding)
         if entry.pathname is None:
-            if isinstance(f, basestring):
+            if isinstance(f, str):
                 st = os.stat(f)
                 entry.pathname = f
                 entry.size = st.st_size
@@ -346,14 +351,17 @@ class Entry(object):
                 entry.pathname = getattr(f, 'pathname', None)
                 entry.size = getattr(f, 'size', 0)
                 entry.mtime = getattr(f, 'mtime', time.time())
-                entry.mode = getattr(f, 'mode', stat.S_IFREG)
+                entry.mode = stat.S_IFREG
         return entry
 
     def to_archive(self, archive):
         '''Creates an archive header and writes it to the given archive.'''
         e = _libarchive.archive_entry_new()
         try:
-            _libarchive.archive_entry_set_pathname(e, self.pathname.encode(self.encoding))
+            if PY3:
+                _libarchive.archive_entry_set_pathname(e, self.pathname)
+            else:
+                _libarchive.archive_entry_set_pathname(e, self.pathname.encode(self.encoding))
             _libarchive.archive_entry_set_filetype(e, stat.S_IFMT(self.mode))
             _libarchive.archive_entry_set_perm(e, stat.S_IMODE(self.mode))
             _libarchive.archive_entry_set_size(e, self.size)
@@ -390,9 +398,9 @@ class Archive(object):
         self._stream = None
         self.encoding = encoding
         self.blocksize = blocksize
-        if isinstance(f, basestring):
+        if isinstance(f, str):
             self.filename = f
-            f = file(f, mode)
+            f = open(f, mode)
             # Only close it if we opened it...
             self._defer_close = True
         elif hasattr(f, 'fileno'):
@@ -520,11 +528,11 @@ class Archive(object):
     def readpath(self, f):
         '''Write current archive entry contents to file. f can be a file-like object or
         a path.'''
-        if isinstance(f, basestring):
+        if isinstance(f, str):
             basedir = os.path.basename(f)
             if not os.path.exists(basedir):
                 os.makedirs(basedir)
-            f = file(f, 'w')
+            f = open(f, 'w')
         return _libarchive.archive_read_data_into_fd(self._a, f.fileno())
 
     def readstream(self, size):
@@ -534,24 +542,34 @@ class Archive(object):
 
     def write(self, member, data=None):
         '''Writes a string buffer to the archive as the given entry.'''
-        if isinstance(member, basestring):
+        if isinstance(member, str):
             member = self.entry_class(pathname=member, encoding=self.encoding)
         if data:
             member.size = len(data)
         member.to_archive(self)
+
         if data:
-            _libarchive.archive_write_data_from_str(self._a, data)
+            if PY3:
+                if isinstance(data, bytes):
+                    result = _libarchive.archive_write_data_from_str(self._a, data)
+                else:
+                    result = _libarchive.archive_write_data_from_str(self._a, data.encode('utf8'))
+            else:
+                result = _libarchive.archive_write_data_from_str(self._a, data)
         _libarchive.archive_write_finish_entry(self._a)
 
-    def writepath(self, f, pathname=None):
+    def writepath(self, f, pathname=None, folder=False):
         '''Writes a file to the archive. f can be a file-like object or a path. Uses
         write() to do the actual writing.'''
         member = self.entry_class.from_file(f, encoding=self.encoding)
-        if isinstance(f, basestring):
+        if isinstance(f, str):
             if os.path.isfile(f):
-                f = file(f, 'r')
+                f = open(f, 'r')
         if pathname:
             member.pathname = pathname
+        if folder and not member.isdir():
+            member.mode = stat.S_IFDIR
+
         if hasattr(f, 'read'):
             # TODO: optimize this to write directly from f to archive.
             self.write(member, data=f.read())
@@ -583,8 +601,8 @@ class SeekableArchive(Archive):
         self._stream = None
         # Convert file to open file. We need this to reopen the archive.
         mode = kwargs.setdefault('mode', 'r')
-        if isinstance(f, basestring):
-            f = file(f, mode)
+        if isinstance(f, str):
+            f = open(f, mode)
         super(SeekableArchive, self).__init__(f, **kwargs)
         self.entries = []
         self.eof = False
@@ -610,7 +628,11 @@ class SeekableArchive(Archive):
     def getentry(self, pathname):
         '''Take a name or entry object and returns an entry object.'''
         for entry in self:
-            if entry.pathname == pathname:
+            if PY3:
+                entry_pathname = entry.pathname
+            if not PY3:
+                entry_pathname = entry.pathname[0]
+            if entry_pathname == pathname:
                 return entry
         raise KeyError(pathname)
 
